.dash-table-container .row {
    display: block;
    margin: 0px;
}

.h1, h1 {
    font-size: 2.5rem;
    text-align: center;
}

.h2, h2 {
    font-size: 2rem;
}

.h3, h3 {
    font-size: 1.75rem;
}

.h4, h4 {
    font-size: 1.5rem;
}

.h5, h5 {
    font-size: 1.25rem;
}

.h6, h6 {
    font-size: 1rem;
}

.centered {
    text-align: center;
}
